from collections import namedtuple


class Price(
    namedtuple(
        "price",
        [
            "quantity",
            "vwap",
            "price",
            "fees",
            "total",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, price):
        return cls(
            quantity=float(price["quantity"]),
            vwap=float(price["vwap"]),
            price=float(price["price"]),
            fees=float(price["fees"]),
            total=float(price["total"]),
            json=price,
        )


class OrderBookEntry(
    namedtuple(
        "book_entry",
        [
            "price",
            "amount",
            "exchange",
        ],
    )
):
    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=float(book_entry[0]),
            amount=float(book_entry[1]),
            exchange=book_entry[2],
        )


class OrderBook(
    namedtuple(
        "order_book",
        [
            "bids",
            "asks",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, order_book):
        return cls(
            bids=[
                OrderBookEntry.create_from_json(book_entry)
                for book_entry in order_book["bids"]
            ],
            asks=[
                OrderBookEntry.create_from_json(book_entry)
                for book_entry in order_book["asks"]
            ],
            json=order_book,
        )



from datetime import datetime

from . import models_v2 as _m
from ..base import Client, ModelMixin


class BitfinexPublic(Client, ModelMixin):
    base_url = "https://api.bitfinex.com/v2/"
    error_keys = ["message"]

    def ticker(self, symbol: str):
        data = self.get(f"ticker/{symbol}")
        if self.return_json:
            return data
        return _m.TradingTicker.create_from_json(data)

    def tickers(self, symbols: list):
        data = self.get(
            "tickers",
            params={
                "symbols": [str(symbol) for symbol in symbols],
            },
        )
        if self.return_json:
            return data
        return {
            ticker[0]: _m.TradingTicker.create_from_json(ticker[1:]) for ticker in data
        }

    def trades(
        self,
        symbol: str,
        limit: int = None,
        start: float = None,
        end: float = None,
        sort: bool = None,
    ):
        if isinstance(start, datetime):
            start = start.timestamp() * 1000
        if isinstance(end, datetime):
            end = end.timestamp() * 1000
        if sort:
            sort = 1 if sort is True else -1
        data = self.get(
            f"trades/{symbol}/hist",
            params={
                "limit": limit,
                "start": start,
                "end": end,
                "sort": sort,
            },
        )
        if self.return_json:
            return data
        return [_m.TradingTrade.create_from_json(trade) for trade in data]

    def books(self, symbol: str, precision: str, length: int = None):
        data = self.get(f"book/{symbol}/{precision}", params={"len": length})
        if self.return_json:
            return data
        return [_m.TradingBook.create_from_json(book) for book in data]

    def stats(
        self,
        symbol: str,
        key: str,
        size: str,
        side: str,
        section: str,
        sort: bool = None,
    ):
        if sort:
            sort = 1 if sort else -1
        data = self.get(
            f"stats1/{key}:{size}:{symbol}:{side}/{section}", params={"sort": sort}
        )
        if self.return_json:
            return data
        if section == "last":
            return _m.Stat.create_from_json(data)
        else:
            return [_m.Stat.create_from_json(stat) for stat in data]

    def stats_last(
        self, symbol: str, key: str, size: str, side: str, sort: bool = None
    ):
        return self.stats(symbol, key, size, side, "last", sort)

    def stats_hist(
        self, symbol: str, key: str, size: str, side: str, sort: bool = None
    ):
        return self.stats(symbol, key, size, side, "hist", sort)

    def candles(
        self,
        symbol: str,
        section: str,
        time_frame: str,
        limit: int = None,
        start: float = None,
        end: float = None,
        sort: bool = None,
    ):
        if isinstance(start, datetime):
            start = start.timestamp() * 1000
        if isinstance(end, datetime):
            end = end.timestamp() * 1000
        if sort:
            sort = 1 if sort else -1
        data = self.get(
            f"candles/trade:{time_frame}:{symbol}/{section}",
            params={
                "limit": limit,
                "start": start,
                "end": end,
                "sort": sort,
            },
        )
        if self.return_json:
            return data
        if section == "last":
            return _m.Candle.create_from_json(data)
        else:
            return [_m.Candle.create_from_json(candle) for candle in data]

    def candles_last(
        self,
        symbol: str,
        time_frame: str,
        limit: int = None,
        start: float = None,
        end: float = None,
        sort: bool = None,
    ):
        return self.candles(symbol, "last", time_frame, limit, start, end, sort)

    def candles_hist(
        self,
        symbol: str,
        time_frame: str,
        limit: int = None,
        start: float = None,
        end: float = None,
        sort: bool = None,
    ):
        return self.candles(symbol, "hist", time_frame, limit, start, end, sort)



from collections import namedtuple
from datetime import datetime


def parse_datetime(datetime_str):
    if datetime_str:
        return datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%fZ")


class Price(
    namedtuple(
        "price",
        [
            "quantity",
            "vwap",
            "price",
            "fees",
            "total",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, price):
        return cls(
            quantity=float(price["quantity"]),
            vwap=float(price["vwap"]),
            price=float(price["price"]),
            fees=float(price["fees"]),
            total=float(price["total"]),
            json=price,
        )


class OrderBookEntry(
    namedtuple(
        "book_entry",
        [
            "price",
            "amount",
            "orders",
        ],
    )
):
    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=book_entry["price"],
            amount=book_entry["amount"],
            orders=book_entry["orders"],
        )


class OrderBook(
    namedtuple(
        "order_book",
        [
            "bids",
            "asks",
            "timestamp",
            "last_price",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, order_book):
        return cls(
            bids=[
                OrderBookEntry.create_from_json(book_entry)
                for book_entry in order_book["bids"]
            ],
            asks=[
                OrderBookEntry.create_from_json(book_entry)
                for book_entry in order_book["asks"]
            ],
            timestamp=order_book["timestamp"],
            last_price=order_book["last_price"],
            json=order_book,
        )


class Trade(
    namedtuple(
        "trade",
        [
            "id",
            "uuid",
            "engine_id",
            "pair",
            "amount",
            "price",
            "created_at",
        ],
    )
):
    @classmethod
    def create_from_json(cls, trade):
        return cls(
            id=trade["id"],
            uuid=trade["uuid"],
            engine_id=trade["engine_id"],
            pair=trade["pair"],
            amount=trade["amount"],
            price=float(trade["price"]),
            created_at=parse_datetime(trade["created_at"]),
        )


class Trades(
    namedtuple(
        "trades",
        [
            "count",
            "next",
            "previous",
            "results",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, trades):
        return cls(
            count=trades["count"],
            next=trades["next"],
            previous=trades["previous"],
            results=[Trade.create_from_json(trade) for trade in trades["results"]],
            json=trades,
        )


class Rates(
    namedtuple(
        "rates",
        [
            "base",
            "rates",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, rates):
        return cls(
            base=rates["base"],
            rates=rates["rates"],
            json=rates,
        )



from ..base import Client


class CoinMarketCap(Client):
    base_url = "https://api.coinmarketcap.com/v1/"
    currencies = None
    timeout = 120

    def ticker(
        self,
        currency: str = None,
        convert: str = None,
        start: int = None,
        limit: int = None,
    ):
        params = {
            "start": start,
            "limit": limit,
            "convert": convert,
        }
        if currency:
            if len(currency) == 3:
                currency = self._get_symbol(currency)["value"]
            data = self.get(f"ticker/{currency}/", params=params)[0]
        else:
            data = self.get("ticker/", params=params)
        return data

    def price(self, currency: str, convert: str = None):
        ticker = self.ticker(currency, convert)
        return float(ticker[f"price_{convert or 'usd'}".lower()])

    def stats(self, convert: str = None):
        data = self.get("global/", params={"convert": convert})
        return data

    def _get_currencies(self):
        ticker = self.ticker()
        return {
            currency["symbol"]: dict(value=currency["id"], decimals=8)
            for currency in ticker
        }

    def _get_symbol(self, currency: str):
        if self.currencies is None:
            self.currencies = self._get_currencies()
        return self.currencies[currency.upper()]



from ..base import _Enum


# Kraken supported symbols
class Symbol(_Enum):
    BCHUSD = "BCHUSD"
    DASHEUR = "DASHEUR"
    DASHUSD = "DASHUSD"
    DASHXBT = "DASHXBT"
    USDTZUSD = "USDTUSD"
    XETCXETH = "ETCETH"
    XETCXXBT = "ETCXBT"
    XETCZEUR = "ETCEUR"
    XETCZUSD = "ETCUSD"
    XETHXXBT = "ETHXBT"
    # XETHXXBT.d = 'ETHXBT.d'
    XETHZCAD = "ETHCAD"
    # XETHZCAD.d = 'ETHCAD.d'
    XETHZEUR = "ETHEUR"
    # XETHZEUR.d = 'ETHEUR.d'
    XETHZGBP = "ETHGBP"
    # XETHZGBP.d = 'ETHGBP.d'
    XETHZJPY = "ETHJPY"
    # XETHZJPY.d = 'ETHJPY.d'
    XETHZUSD = "ETHUSD"
    # XETHZUSD.d = 'ETHUSD.d'
    XICNXETH = "ICNETH"
    XICNXXBT = "ICNXBT"
    XLTCXXBT = "LTCXBT"
    XLTCZEUR = "LTCEUR"
    XLTCZUSD = "LTCUSD"
    XMLNXETH = "MLNETH"
    XMLNXXBT = "MLNXBT"
    XREPXETH = "REPETH"
    XREPXXBT = "REPXBT"
    XREPZEUR = "REPEUR"
    XREPZUSD = "REPUSD"
    XXBTZCAD = "XBTCAD"
    # XXBTZCAD.d = 'XBTCAD.d'
    XXBTZEUR = "XBTEUR"
    # XXBTZEUR.d = 'XBTEUR.d'
    XXBTZGBP = "XBTGBP"
    # XXBTZGBP.d = 'XBTGBP.d'
    XXBTZJPY = "XBTJPY"
    # XXBTZJPY.d = 'XBTJPY.d'
    XXBTZUSD = "XBTUSD"
    # XXBTZUSD.d = 'XBTUSD.d'
    XXDGXXBT = "XDGXBT"
    XXLMXXBT = "XLMXBT"
    XXLMZEUR = "XLMEUR"
    XXLMZUSD = "XLMUSD"
    XXMRXXBT = "XMRXBT"
    XXMRZEUR = "XMREUR"
    XXMRZUSD = "XMRUSD"
    XXRPXXBT = "XRPXBT"
    XZECXXBT = "ZECXBT"
    XZECZEUR = "ZECEUR"
    XZECZUSD = "ZECUSD"


# Kraken supported currencies
class Currency(_Enum):
    DASH = "DASH"
    KFEE = "FEE"
    USDT = "USDT"
    XDAO = "DAO"
    XETC = "ETC"
    XETH = "ETH"
    XICN = "ICN"
    XLTC = "LTC"
    XMLN = "MLN"
    XNMC = "NMC"
    XREP = "REP"
    XXBT = "XBT"
    XXDG = "XDG"
    XXLM = "XLM"
    XXMR = "XMR"
    XXRP = "XRP"
    XXVN = "XVN"
    XZEC = "ZEC"
    ZCAD = "CAD"
    ZEUR = "EUR"
    ZGBP = "GBP"
    ZJPY = "JPY"
    ZKRW = "KRW"
    ZUSD = "USD"



from ..auth import ApiKeyAuth
from ..base import AuthMixin, Client
from ..common import format_date_iso, format_datetime_iso


class OXR(Client, AuthMixin):
    base_url = "https://openexchangerates.org/api/"
    error_keys = ["error"]

    def __init__(self, app_id: str, timeout: int = None, **kwargs):
        super().__init__(timeout, **kwargs)
        self.auth = ApiKeyAuth(app_id, api_key_param="app_id")

    def currencies(self):
        """
        Get a JSON list of all currency symbols available from the Open
        Exchange Rates API, along with their full names.
        ref. https://oxr.readme.io/docs/currencies-json
        """
        return self.get("currencies.json")

    def latest(self, base: str = None, symbols: list = None):
        """
        Get latest data.
        ref. https://oxr.readme.io/docs/latest-json
        """
        return self.get("latest.json", base, symbols)

    def historical(self, date_for: str, base: str = None, symbols: list = None):
        """
        Get daily historical data
        ref. https://oxr.readme.io/docs/historical-json
        """
        date_for = format_date_iso(date_for)
        return self.get(f"historical/{date_for}.json", base, symbols)

    def time_series(self, start: str, end: str, base: str = None, symbols: list = None):
        """
        Get time-series data.
        ref. https://oxr.readme.io/docs/time-series-json
        """
        return self.get(
            "time-series.json",
            base,
            symbols,
            params={
                "start": format_date_iso(start),
                "end": format_date_iso(end),
            },
        )

    def convert(self, value: int, from_symbol: str, to_symbol: str):
        """
        Convert any money value from one currency to another at the latest
        API rates.
        ref. https://oxr.readme.io/docs/convert
        """
        return self.get(f"convert/{value}/{from_symbol}/{to_symbol}")

    def ohlc(
        self, start_time: str, period: str, symbols: list = None, base: str = None
    ):
        """
        Get historical open, high, low, close (OHLC) and average data.
        ref. https://oxr.readme.io/docs/ohlc-json
        """
        return self.get(
            "ohlc.json",
            base,
            symbols,
            params={
                "start_time": format_datetime_iso(start_time),
                "period": period,
            },
        )

    # OVERRIDES ---------------------------------------------------------------
    def get(
        self, endpoint: str, base: str = None, symbols: list = None, params: dict = None
    ):
        params = params or {}
        if base is not None:
            params["base"] = base
        if isinstance(symbols, list) or isinstance(symbols, tuple):
            symbols = ",".join(symbols)
        if symbols is not None:
            params["symbols"] = symbols
        return super().get(endpoint, params=params)



from ..base import _Enum


# Bitfinex supported symbols
class Symbol(_Enum):
    BCHUSD = "tBCHUSD"
    BCHBTC = "tBCHBTC"
    BTCUSD = "tBTCUSD"
    LTCUSD = "tLTCUSD"
    ETHUSD = "tETHUSD"
    ETHBTC = "tETHBTC"
    ETCUSD = "tETCUSD"
    ETCBTC = "tETCBTC"
    BFXUSD = "tBFXUSD"
    BFXBTC = "tBFXBTC"
    RRTUSD = "tRRTUSD"
    RRTBTC = "tRRTBTC"
    ZECUSD = "tZECUSD"
    ZECBTC = "tZECBTC"
    XMRUSD = "tXMRUSD"
    XMRBTC = "tXMRBTC"


# Bitfinex supported currencies
class Currency(_Enum):
    BCH = "bch"
    BTC = "btc"
    ETH = "eth"
    USD = "usd"


# Bitfinex supported precisions
class BookPrecision(_Enum):
    P0 = "P0"
    P1 = "P1"
    P2 = "P2"
    P3 = "P3"
    R0 = "R0"



from ..base import Client


class KrakenPublic(Client):
    base_url = "https://api.kraken.com/0/"
    error_keys = ["error"]

    def server_time(self):
        return self.get("public/Time")

    def assets(self):
        return self.get("public/Assets")

    def asset_pairs(self):
        return self.get("public/AssetPairs")

    def ticker(self, symbol: str):
        return self.get(
            "public/Ticker",
            params={
                "pair": str(symbol),
            },
        )

    def ohlc(self, symbol: str, interval: int = None, since: str = None):
        return self.get(
            "public/OHLC",
            params={
                "pair": str(symbol),
                "interval": interval,
                "since": since,
            },
        )

    def order_book(self, symbol: str, count: int = None):
        return self.get(
            "public/Depth",
            params={
                "pair": str(symbol),
                "count": count,
            },
        )

    def trades(self, symbol: str, since: str = None):
        return self.get(
            "public/Trades",
            params={
                "pair": str(symbol),
                "since": since,
            },
        )

    def spread(self, symbol: str, since: str = None):
        return self.get(
            "public/Spread",
            params={
                "pair": str(symbol),
                "since": since,
            },
        )



from datetime import datetime, timedelta

from ..base import Client
from ..common import current_utc_date, date_range


class CoinDesk(Client):
    base_url = "http://api.coindesk.com/v1/"
    timeout = 15

    def bpi(self, currency: str):
        return _BPI(self, currency)

    def rate(self, currency: str):
        return _Rate(self, currency)


class _BPI(CoinDesk):
    def __init__(self, parent, currency: str):
        super().__init__(timeout=parent.timeout)
        self.currency = currency.upper()

    def current(self):
        return self.get(f"bpi/currentprice/{self.currency}.json")

    def historical(
        self, start: datetime, end: datetime = None, include_today: bool = False
    ):
        if isinstance(start, datetime):
            start = start.date()
        if isinstance(end, datetime):
            end = end.date()
        today = current_utc_date()
        end = end or today
        # Validate dates
        assert start <= end, "'start' date must <= 'end' date'!"
        self._validate_historical_date(start)
        self._validate_historical_date(end)
        # If start date is today, return only current BPI
        if start == today:
            current = self.bpi(self.currency).current()
            current["bpi"] = {str(today): current["bpi"][self.currency]["rate_float"]}
            return current
        # Normal call for historical BPI
        response = self.get(
            "bpi/historical/close.json",
            params={
                "currency": self.currency,
                "start": start,
                "end": end,
            },
        )
        # If end date is today, add current BPI
        if end == today and include_today:
            response["bpi"][str(today)] = self.rate(self.currency).current()
        # Validate response
        historical_bpi = response["bpi"]
        for d in date_range(start, end):
            assert historical_bpi[str(d)], f"{d} is not present in BPI!"

        return response

    @staticmethod
    def _validate_historical_date(date):
        if not date:
            return
        current_date = current_utc_date()
        msg = f"({date}) must be a date <= current date " f"({current_date})"
        assert date <= current_date, msg


class _Rate(CoinDesk):
    def __init__(self, parent, currency: str):
        super().__init__(timeout=parent.timeout)
        self.currency = currency.upper()
        self._bpi = self.bpi(self.currency)

    def current(self):
        response = self._bpi.current()
        rate = response["bpi"][self.currency]["rate_float"]
        return rate

    def historical(
        self, start: datetime, end: datetime = None, include_today: bool = False
    ):
        response = self._bpi.historical(start, end, include_today)
        rate_dict = response["bpi"]
        return rate_dict

    def for_date(self, date_for: datetime):
        if isinstance(date_for, datetime):
            date_for = date_for.date()
        rate_dict = self.historical(start=date_for, end=date_for)
        rate = rate_dict[str(date_for)]
        return rate

    def since_date(self, date_since: datetime, include_today: bool = False):
        if isinstance(date_since, datetime):
            date_since = date_since.date()
        rate_dict = self.historical(date_since, None, include_today)
        return rate_dict

    def last_n_days(self, n_days: int, include_today: bool = False):
        start = current_utc_date() - timedelta(days=n_days)
        rate_dict = self.historical(start, None, include_today)
        return rate_dict



from ..base import _Enum, Currency, Market

# Limits
ORDERS_LIMIT = 300
TRANSFERS_LIMIT = 300


class Currency(Currency):
    ARS = dict(value="ARS")
    BCH = dict(value="BCH", decimals=8)
    BTC = dict(value="BTC", decimals=8)
    CLP = dict(value="CLP")
    COP = dict(value="COP")
    ETH = dict(value="ETH", decimals=9)
    LTC = dict(value="LTC", decimals=8)
    PEN = dict(value="PEN")
    USDC = dict(value="USDC")


class Market(Market):
    BCH_ARS = dict(value="BCH-ARS", base=Currency.BCH, quote=Currency.ARS)
    BCH_BTC = dict(value="BCH-BTC", base=Currency.BCH, quote=Currency.BTC)
    BCH_CLP = dict(value="BCH-CLP", base=Currency.BCH, quote=Currency.CLP)
    BCH_COP = dict(value="BCH-COP", base=Currency.BCH, quote=Currency.COP)
    BTC_ARS = dict(value="BTC-ARS", base=Currency.BTC, quote=Currency.ARS)
    BCH_PEN = dict(value="BCH-PEN", base=Currency.BCH, quote=Currency.PEN)
    BTC_CLP = dict(value="BTC-CLP", base=Currency.BTC, quote=Currency.CLP)
    BTC_COP = dict(value="BTC-COP", base=Currency.BTC, quote=Currency.COP)
    BTC_PEN = dict(value="BTC-PEN", base=Currency.BTC, quote=Currency.PEN)
    ETH_ARS = dict(value="ETH-ARS", base=Currency.ETH, quote=Currency.ARS)
    ETH_BTC = dict(value="ETH-BTC", base=Currency.ETH, quote=Currency.BTC)
    ETH_CLP = dict(value="ETH-CLP", base=Currency.ETH, quote=Currency.CLP)
    ETH_COP = dict(value="ETH-COP", base=Currency.ETH, quote=Currency.COP)
    ETH_PEN = dict(value="ETH-PEN", base=Currency.ETH, quote=Currency.PEN)
    LTC_ARS = dict(value="LTC-ARS", base=Currency.LTC, quote=Currency.ARS)
    LTC_BTC = dict(value="LTC-BTC", base=Currency.LTC, quote=Currency.BTC)
    LTC_CLP = dict(value="LTC-CLP", base=Currency.LTC, quote=Currency.CLP)
    LTC_COP = dict(value="LTC-COP", base=Currency.LTC, quote=Currency.COP)
    LTC_PEN = dict(value="LTC-PEN", base=Currency.LTC, quote=Currency.PEN)
    BTC_USDC = dict(value="BTC-USDC", base=Currency.BTC, quote=Currency.USDC)
    USDC_ARS = dict(value="USDC-ARS", base=Currency.USDC, quote=Currency.ARS)
    USDC_CLP = dict(value="USDC-CLP", base=Currency.USDC, quote=Currency.CLP)
    USDC_COP = dict(value="USDC-COP", base=Currency.USDC, quote=Currency.COP)
    USDC_PEN = dict(value="USDC-PEN", base=Currency.USDC, quote=Currency.PEN)


class QuotationType(_Enum):
    BID_GIVEN_SIZE = "bid_given_size"
    BID_GIVEN_EARNED_BASE = "bid_given_earned_base"
    BID_GIVEN_SPENT_QUOTE = "bid_given_spent_quote"
    ASK_GIVEN_SIZE = "ask_given_size"
    ASK_GIVEN_EARNED_QUOTE = "ask_given_earned_quote"
    ASK_GIVEN_SPENT_BASE = "ask_given_spent_base"


class OrderType(_Enum):
    ASK = "Ask"
    BID = "Bid"


class OrderPriceType(_Enum):
    MARKET = "market"
    LIMIT = "limit"


class OrderState(_Enum):
    RECEIVED = "received"
    PENDING = "pending"
    TRADED = "traded"
    CANCELING = "canceling"
    CANCELED = "canceled"


class BalanceEvent(_Enum):
    DEPOSIT_CONFIRM = "deposit_confirm"
    WITHDRAWAL_CONFIRM = "withdrawal_confirm"
    TRANSACTION = "transaction"
    TRANSFER_CONFIRMATION = "transfer_confirmation"


class ReportType(_Enum):
    CANDLESTICK = "candlestick"
    AVERAGE_PRICES = "average_prices"



import base64

from requests import PreparedRequest as P

from . import constants as _c
from . import models as _m
from .client_public import BudaPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class BudaHMACAuth(HMACAuth):

    api_key_header = "X-SBTC-APIKEY"
    nonce_header = "X-SBTC-NONCE"
    signature_header = "X-SBTC-SIGNATURE"
    signature_delimiter = " "
    algorithm = "sha384"

    def build_message(self, r: P, nonce: str):
        components = [r.method, r.path_url]
        if r.body:
            encoded_body = base64.b64encode(r.body).decode()
            components.append(encoded_body)
        components.append(nonce)
        message = self.signature_delimiter.join(components)
        return message


class BudaAuth(BudaPublic, AuthMixin):
    auth_cls = BudaHMACAuth

    def __init__(self, key: str, secret: str, timeout: int = None, **kwargs):
        super().__init__(timeout, **kwargs)
        self.add_auth(key, secret)

    # BALANCES-----------------------------------------------------------------
    def balance(self, currency: str):
        data = self.get(f"balances/{currency}")
        if self.return_json:
            return data
        return _m.Balance.create_from_json(data["balance"])

    def balance_event_pages(
        self,
        currencies: list,
        event_names: list,
        page: int = None,
        per_page: int = None,
        relevant: bool = None,
    ):
        data = self.get(
            "balance_events",
            params={
                "currencies[]": [str(c) for c in currencies],
                "event_names[]": [str(e) for e in event_names],
                "page": page,
                "per": per_page,
                "relevant": relevant,
            },
        )
        if self.return_json:
            return data
        # TODO: Response only contains a 'total_count' field instead of meta
        return _m.BalanceEventPages.create_from_json(
            data["balance_events"], data["total_count"], page
        )

    # ORDERS ------------------------------------------------------------------
    def new_order_payload(self, market_id: str, payload):
        data = self.post(f"markets/{market_id}/orders", json=payload)
        if self.return_json:
            return data
        return _m.Order.create_from_json(data["order"])

    def new_order(
        self,
        market_id: str,
        order_type: str,
        price_type: str,
        amount: float,
        limit: float = None,
    ):
        return self.new_order_payload(
            market_id,
            payload={
                "type": str(order_type),
                "price_type": str(price_type),
                "amount": str(amount),
                "limit": str(limit) if limit else None,
            },
        )

    def order_pages(
        self,
        market_id: str,
        page: int = None,
        per_page: int = None,
        state: str = None,
        minimum_exchanged: float = None,
    ):
        if per_page and per_page > _c.ORDERS_LIMIT:
            msg = f"Param 'per_page' must be <= {_c.ORDERS_LIMIT}"
            raise ValueError(msg)
        data = self.get(
            f"markets/{market_id}/orders",
            params={
                "per": per_page,
                "page": page,
                "state": str(state) if state else None,
                "minimum_exchanged": minimum_exchanged,
            },
        )
        if self.return_json:
            return data
        return _m.OrderPages.create_from_json(data["orders"], data.get("meta"))

    def batch_orders(self, cancel_list: list = None, place_list: list = None):
        diff = {"diff": []}
        if cancel_list:
            for idx in cancel_list:
                diff["diff"].append({"mode": "cancel", "order_id": idx})
        if place_list:
            for order in place_list:
                diff["diff"].append({"mode": "place", "order": order})
        data = self.post("orders", json=diff)
        return data

    def order_details(self, order_id: int):
        data = self.get(f"orders/{order_id}")
        if self.return_json:
            return data
        return _m.Order.create_from_json(data["order"])

    def cancel_order(self, order_id: int):
        data = self.put(
            f"orders/{order_id}",
            json={
                "state": _c.OrderState.CANCELING.value,
            },
        )
        if self.return_json:
            return data
        return _m.Order.create_from_json(data["order"])

    # PAYMENTS ----------------------------------------------------------------
    def _transfers(
        self,
        path,
        model,
        key,
        page: int = None,
        per_page: int = None,
        state: str = None,
    ):
        if per_page and per_page > _c.TRANSFERS_LIMIT:
            msg = f"Param 'per_page' must be <= {_c.TRANSFERS_LIMIT}"
            raise ValueError(msg)
        data = self.get(
            path,
            params={
                "per": per_page,
                "page": page,
                "state": str(state) if state else None,
            },
        )
        if self.return_json:
            return data
        return model.create_from_json(data[key], data.get("meta"))

    def withdrawal_pages(self, currency: str, page: int = None, per_page: int = None):
        return self._transfers(
            path=f"currencies/{currency}/withdrawals",
            model=_m.WithdrawalPages,
            key="withdrawals",
            page=page,
            per_page=per_page,
        )

    def withdrawals(self, currency: str, page: int = None, per_page: int = None):
        data = self.withdrawal_pages(currency, page, per_page)
        if isinstance(data, dict):
            return data["withdrawals"]
        return data.withdrawals

    def deposit_pages(self, currency: str, page: int = None, per_page: int = None):
        return self._transfers(
            path=f"currencies/{currency}/deposits",
            model=_m.DepositPages,
            key="deposits",
            page=page,
            per_page=per_page,
        )

    def deposits(self, currency: str, page: int = None, per_page: int = None):
        data = self.deposit_pages(currency, page, per_page)
        if isinstance(data, dict):
            return data["deposits"]
        return data.deposits

    def withdrawal(
        self,
        currency: str,
        amount: float,
        target_address: str = None,
        amount_includes_fee: bool = True,
        simulate: bool = False,
    ):
        data = self.post(
            f"currencies/{currency}/withdrawals",
            json={
                "withdrawal_data": {
                    "target_address": target_address,
                },
                "amount": str(amount),
                "currency": str(currency),
                "simulate": simulate,
                "amount_includes_fee": amount_includes_fee,
            },
        )
        if self.return_json:
            return data
        return _m.Withdrawal.create_from_json(data["withdrawal"])

    def simulate_withdrawal(
        self, currency: str, amount: float, amount_includes_fee: bool = True
    ):
        return self.withdrawal(
            currency,
            amount,
            target_address=None,
            amount_includes_fee=amount_includes_fee,
            simulate=True,
        )



from ..base import _Enum


class CurrencyPair(_Enum):
    BTC_USD = "btcusd"
    BTC_EUR = "btceur"
    EUR_USD = "eurusd"
    XRP_USD = "xrpusd"
    XRP_EUR = "xrpeur"
    XRP_BTC = "xrpbtc"
    LTC_USD = "ltcusd"
    LTC_EUR = "ltceur"
    LTC_BTC = "ltcbtc"
    ETH_USD = "ethusd"
    ETH_EUR = "etheur"
    ETH_BTC = "ethbtc"
    BCH_USD = "bchusd"
    BCH_EUR = "bcheur"
    BCH_BTC = "bchbtc"


class TimeInterval(_Enum):
    DAY = "day"
    HOUR = "hour"
    MINUTE = "minute"



from .bitcoinity import *  # noqa: F401
from .bitex import *  # noqa: F401
from .bitfinex import *  # noqa: F401
from .bitstamp import *  # noqa: F401
from .buda import *  # noqa: F401
from .coindesk import *  # noqa: F401
from .coinmarketcap import *  # noqa: F401
from .cryptomkt import *  # noqa: F401
from .currencylayer import *  # noqa: F401
from .errors import *  # noqa: F401
from .kraken import *  # noqa: F401
from .oxr import *  # noqa: F401
from .ripio import *  # noqa: F401
from .sfox import *  # noqa: F401
from .surbtc import *  # noqa: F401
from ._version import __version__  # noqa: F40



from datetime import datetime

from . import constants as _c
from . import models as _m
from ..base import Client, ModelMixin


class BudaPublic(Client, ModelMixin):
    base_url = "https://www.buda.com/api/v2/"
    error_keys = ["message"]

    def markets(self):
        data = self.get("markets")
        if self.return_json:
            return data
        return [_m.Market.create_from_json(market) for market in data["markets"]]

    def market_details(self, market_id: str):
        data = self.get(f"markets/{market_id}")
        if self.return_json:
            return data
        return _m.Market.create_from_json(data["market"])

    def ticker(self, market_id: str):
        data = self.get(f"markets/{market_id}/ticker")
        if self.return_json:
            return data
        return _m.Ticker.create_from_json(data["ticker"])

    def order_book(self, market_id: str):
        data = self.get(f"markets/{market_id}/order_book")
        if self.return_json:
            return data
        return _m.OrderBook.create_from_json(data["order_book"])

    def trades(self, market_id: str, timestamp: int = None, limit: int = None):
        data = self.get(
            f"markets/{market_id}/trades",
            params={
                "timestamp": timestamp,
                "limit": limit,
            },
        )
        if self.return_json:
            return data
        return _m.Trades.create_from_json(data["trades"])

    def quotation(
        self, market_id: str, quotation_type: str, amount: float, limit: float = None
    ):
        data = self.post(
            f"markets/{market_id}/quotations",
            json={
                "quotation": {
                    "type": str(quotation_type),
                    "amount": str(amount),
                    "limit": str(limit) if limit else None,
                },
            },
        )
        if self.return_json:
            return data
        return _m.Quotation.create_from_json(data["quotation"])

    def quotation_market(self, market_id: str, quotation_type: str, amount: float):
        return self.quotation(market_id, quotation_type, amount, limit=None)

    def quotation_limit(
        self, market_id: str, quotation_type: str, amount: float, limit: float
    ):
        return self.quotation(market_id, quotation_type, amount, limit)

    def _report(
        self,
        market_id: str,
        report_type: _c.ReportType,
        start_at: datetime = None,
        end_at: datetime = None,
    ):
        if isinstance(start_at, datetime):
            start_at = int(start_at.timestamp())
        if isinstance(end_at, datetime):
            end_at = int(end_at.timestamp())
        data = self.get(
            f"markets/{market_id}/reports",
            params={
                "report_type": str(report_type),
                "from": start_at,
                "to": end_at,
            },
        )
        return data

    def report_average_prices(
        self, market_id: str, start_at: datetime = None, end_at: datetime = None
    ):
        data = self._report(market_id, _c.ReportType.AVERAGE_PRICES, start_at, end_at)
        if self.return_json:
            return data
        return [_m.AveragePrice.create_from_json(report) for report in data["reports"]]

    def report_candlestick(
        self, market_id: str, start_at: datetime = None, end_at: datetime = None
    ):
        data = self._report(market_id, _c.ReportType.CANDLESTICK, start_at, end_at)
        if self.return_json:
            return data
        return [_m.Candlestick.create_from_json(report) for report in data["reports"]]



from collections import namedtuple


class TradingTicker(
    namedtuple(
        "trading_ticker",
        [
            "BID",
            "BID_SIZE",
            "ASK",
            "ASK_SIZE",
            "DAILY_CHANGE",
            "DAILY_CHANGE_PERC",
            "LAST_PRICE",
            "VOLUME",
            "HIGH",
            "LOW",
        ],
    )
):
    @classmethod
    def create_from_json(cls, ticker):
        return cls(
            BID=ticker[0],
            BID_SIZE=ticker[1],
            ASK=ticker[2],
            ASK_SIZE=ticker[3],
            DAILY_CHANGE=ticker[4],
            DAILY_CHANGE_PERC=ticker[5],
            LAST_PRICE=ticker[6],
            VOLUME=ticker[7],
            HIGH=ticker[8],
            LOW=ticker[9],
        )


class FoundingTicker(
    namedtuple(
        "founding_ticker",
        [
            "FRR",
            "BID",
            "BID_SIZE",
            "BID_PERIOD",
            "ASK",
            "ASK_SIZE",
            "ASK_PERIOD",
            "DAILY_CHANGE",
            "DAILY_CHANGE_PERC",
            "LAST_PRICE",
            "VOLUME",
            "HIGH",
            "LOW",
        ],
    )
):
    @classmethod
    def create_from_json(cls, ticker):
        return cls(
            FRR=ticker[0],
            BID=ticker[1],
            BID_SIZE=ticker[2],
            BID_PERIOD=ticker[3],
            ASK=ticker[4],
            ASK_SIZE=ticker[5],
            ASK_PERIOD=ticker[6],
            DAILY_CHANGE=ticker[7],
            DAILY_CHANGE_PERC=ticker[8],
            LAST_PRICE=ticker[9],
            VOLUME=ticker[10],
            HIGH=ticker[11],
            LOW=ticker[12],
        )


class TradingTrade(
    namedtuple(
        "trading_trade",
        [
            "ID",
            "MTS",
            "AMOUNT",
            "PRICE",
        ],
    )
):
    @classmethod
    def create_from_json(cls, trade):
        return cls(
            ID=trade[0],
            MTS=trade[1],
            AMOUNT=trade[2],
            PRICE=trade[3],
        )


class FoundingTrade(
    namedtuple(
        "founding_trade",
        [
            "ID",
            "MTS",
            "AMOUNT",
            "RATE",
            "PERIOD",
        ],
    )
):
    @classmethod
    def create_from_json(cls, trade):
        return cls(
            ID=trade[0],
            MTS=trade[1],
            AMOUNT=trade[2],
            RATE=trade[3],
            PERIOD=trade[4],
        )


class TradingBook(
    namedtuple(
        "trading_book",
        [
            "PRICE",
            "COUNT",
            "AMOUNT",
        ],
    )
):
    @classmethod
    def create_from_json(cls, book):
        return cls(
            PRICE=book[0],
            COUNT=book[1],
            AMOUNT=book[2],
        )


class FoundingBook(
    namedtuple(
        "founding_book",
        [
            "RATE",
            "PERIOD",
            "COUNT",
            "AMOUNT",
        ],
    )
):
    @classmethod
    def create_from_json(cls, book):
        return cls(
            RATE=book[0],
            PERIOD=book[1],
            COUNT=book[2],
            AMOUNT=book[3],
        )


class Stat(
    namedtuple(
        "stat",
        [
            "MTS",
            "VALUE",
        ],
    )
):
    @classmethod
    def create_from_json(cls, stat):
        return cls(
            MTS=stat[0],
            VALUE=stat[1],
        )


class Candle(
    namedtuple(
        "candle",
        [
            "MTS",
            "OPEN",
            "CLOSE",
            "HIGH",
            "LOW",
            "VOLUME",
        ],
    )
):
    @classmethod
    def create_from_json(cls, candle):
        return cls(
            MTS=candle[0],
            OPEN=candle[1],
            CLOSE=candle[2],
            HIGH=candle[3],
            LOW=candle[4],
            VOLUME=candle[5],
        )



import json
from urllib.parse import urlsplit

from requests import PreparedRequest as P

from . import constants as _c
from . import models as _m
from .client_public import CryptoMKTPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class CryptoMKTHMACAuth(HMACAuth):

    api_key_header = "X-MKT-APIKEY"
    nonce_header = "X-MKT-TIMESTAMP"
    signature_header = "X-MKT-SIGNATURE"
    signature_delimiter = ""
    algorithm = "sha384"

    def _nonce(self):
        return self.timestamp.seconds()

    def build_message(self, r: P, nonce: str):
        path = urlsplit(r.path_url).path
        body = json.loads(r.body) if r.body else {}
        components = [nonce, path]
        for key in sorted(body.keys()):
            value = str(body[key])
            components.append(value)
        message = self.signature_delimiter.join(components)
        return message


class CryptoMKTAuth(CryptoMKTPublic, AuthMixin):
    auth_cls = CryptoMKTHMACAuth

    def __init__(self, key: str, secret: str, timeout: int = None, **kwargs):
        super().__init__(timeout, **kwargs)
        self.add_auth(key, secret)

    # BALANCE------------------------------------------------------------------
    def balance(self):
        data = self.get("balance")
        if self.return_json:
            return data
        return _m.Balance.create_from_json(data["data"])

    def wallet_balance(self, currency: str):
        balance = self.balance()
        return getattr(balance, str(currency))

    # ORDERS-------------------------------------------------------------------
    def active_orders(
        self, market_id: str, page: int = None, limit: int = _c.ORDERS_LIMIT
    ):
        data = self.get(
            "orders/active",
            params={
                "market": str(market_id),
                "page": page,
                "limit": limit,
            },
        )
        if self.return_json:
            return data
        return _m.Orders.create_from_json(data["data"], data["pagination"])

    def executed_orders(
        self, market_id: str, page: int = None, limit: int = _c.ORDERS_LIMIT
    ):
        data = self.get(
            "orders/executed",
            params={
                "market": str(market_id),
                "page": page,
                "limit": limit,
            },
        )
        if self.return_json:
            return data
        return _m.Orders.create_from_json(data["data"], data["pagination"])

    def create_order(
        self, market_id: str, order_type: str, amount: float, price: float
    ):
        data = self.post(
            "orders/create",
            json={
                "market": str(market_id),
                "type": str(order_type),
                "amount": amount,
                "price": price,
            },
        )
        if self.return_json:
            return data
        return _m.Order.create_from_json(data["data"])

    def order_status(self, order_id: str):
        data = self.get(
            "orders/status",
            params={
                "id": order_id,
            },
        )
        if self.return_json:
            return data
        return _m.Order.create_from_json(data["data"])

    def cancel_order(self, order_id: str):
        data = self.get(
            "orders/cancel",
            params={
                "id": order_id,
            },
        )
        if self.return_json:
            return data
        return _m.Order.create_from_json(data["data"])

    # PAYMENTS-----------------------------------------------------------------
    # TODO: Not tested
    def create_payment(
        self,
        amount: float,
        currency: str,
        account_email: str,
        external_id: str = None,
        callback_url: str = None,
        error_url: str = None,
        success_url: str = None,
        refund_email: str = None,
    ):
        data = self.post(
            "payment/new_order",
            json={
                "to_receive": amount,
                "to_receive_currency": str(currency),
                "payment_receiver": account_email,
                "external_id": external_id,
                "callback_url": callback_url,
                "error_url": error_url,
                "success_url": success_url,
                "refund_email": refund_email,
            },
        )
        if self.return_json:
            return data
        return data

    # TODO: Not tested
    def payment_status(self, payment_id: str):
        data = self.get(
            "payment/status",
            params={
                "id": payment_id,
            },
        )
        return data



from requests import Response

from requests.exceptions import RequestException


class APIException(RequestException):
    pass


class InvalidResponse(APIException):
    def __init__(self, error_msg: str, r: Response):
        # Add status code
        message = str(r.status_code)
        # Check for client or server errors
        if 400 <= r.status_code < 500:
            message = f"{message} Client Error:"
        elif 500 <= r.status_code < 600:
            message = f"{message} Server Error:"
        # Add reason
        message = f"{message} {r.reason}"
        # Add message from source
        if error_msg:
            message = f"{message} ({error_msg})"
        # Add url
        message = f"{message} for url: {r.url}"

        super().__init__(message, response=r)
        self.message = message


class DecodeError(APIException):
    def __init__(self, msg, r: Response):
        super().__init__(msg, response=r)



from ..base import _Enum


# Bitfinex supported symbols
class Symbol(_Enum):
    BTCUSD = "btcusd"
    LTCUSD = "ltcusd"
    ETHUSD = "ethusd"
    ETHBTC = "ethbtc"
    ETCUSD = "etcusd"
    ETCBTC = "etcbtc"
    BFXUSD = "bfxusd"
    BFXBTC = "bfxbtc"
    RRTUSD = "rrtusd"
    RRTBTC = "rrtbtc"
    ZECUSD = "zecusd"
    ZECBTC = "zecbtc"
    XMRUSD = "xmrusd"
    XMRBTC = "xmrbtc"
    BCHUSD = "bchusd"
    BCHBTC = "bchbtc"
    BCHETH = "bcheth"


# Bitfinex supported currencies
class Currency(_Enum):
    USD = "usd"
    BTC = "btc"
    ETH = "eth"
    BCH = "bch"
    BFX = "bfx"



from collections import namedtuple


class Ticker(
    namedtuple(
        "ticker",
        [
            "last",
            "price_before_last",
            "open",
            "high",
            "low",
            "vwap",
            "volume",
            "bid",
            "ask",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, ticker):
        return cls(
            last=ticker["last"],
            price_before_last=ticker["price_before_last"],
            open=ticker["open"],
            high=ticker["high"],
            low=ticker["low"],
            vwap=ticker["vwap"],
            volume=ticker["volume"],
            bid=ticker["bid"],
            ask=ticker["ask"],
            json=ticker,
        )


class OrderBookEntry(
    namedtuple(
        "book_entry",
        [
            "price",
            "amount",
        ],
    )
):
    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=float(book_entry[0]),
            amount=float(book_entry[1]),
        )


class OrderBook(
    namedtuple(
        "order_book",
        [
            "asks",
            "bids",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, order_book):
        return cls(
            asks=[
                OrderBookEntry.create_from_json(entry) for entry in order_book["asks"]
            ],
            bids=[
                OrderBookEntry.create_from_json(entry) for entry in order_book["bids"]
            ],
            json=order_book,
        )


class Transaction(
    namedtuple(
        "transaction",
        [
            "timestamp",
            "id",
            "price",
            "amount",
        ],
    )
):
    @classmethod
    def create_from_json(cls, transaction):
        return cls(
            timestamp=transaction[0],
            id=transaction[1],
            price=transaction[2],
            amount=transaction[3],
        )



from ..base import _Enum, Market, Currency

# Limits
ORDERS_LIMIT = 100


class Currency(Currency):
    ARS = dict(value="ARS")
    BRL = dict(value="BRL")
    BTC = dict(value="BTC")
    CLP = dict(value="CLP")
    EOS = dict(value="EOS")
    ETH = dict(value="ETH", decimals=9)
    EUR = dict(value="EUR")
    MXN = dict(value="MXN")
    XLM = dict(value="XLM")


class OrderType(_Enum):
    BUY = "buy"
    SELL = "sell"


class Market(Market):
    BTC_ARS = dict(value="BTCARS", base=Currency.BTC, quote=Currency.ARS)
    BTC_BRL = dict(value="BTCBRL", base=Currency.BTC, quote=Currency.BRL)
    BTC_CLP = dict(value="BTCCLP", base=Currency.BTC, quote=Currency.CLP)
    BTC_EUR = dict(value="BTCEUR", base=Currency.BTC, quote=Currency.EUR)
    BTC_MXN = dict(value="BTCMXN", base=Currency.BTC, quote=Currency.MXN)
    EOS_ARS = dict(value="EOSARS", base=Currency.EOS, quote=Currency.ARS)
    EOS_BRL = dict(value="EOSBRL", base=Currency.EOS, quote=Currency.BRL)
    EOS_CLP = dict(value="EOSCLP", base=Currency.EOS, quote=Currency.CLP)
    EOS_EUR = dict(value="EOSEUR", base=Currency.EOS, quote=Currency.EUR)
    EOS_MXN = dict(value="EOSMXN", base=Currency.EOS, quote=Currency.MXN)
    ETH_ARS = dict(value="ETHARS", base=Currency.ETH, quote=Currency.ARS)
    ETH_BRL = dict(value="ETHBRL", base=Currency.ETH, quote=Currency.BRL)
    ETH_CLP = dict(value="ETHCLP", base=Currency.ETH, quote=Currency.CLP)
    ETH_EUR = dict(value="ETHEUR", base=Currency.ETH, quote=Currency.EUR)
    ETH_MXN = dict(value="ETHMXN", base=Currency.ETH, quote=Currency.MXN)
    XLM_ARS = dict(value="XLMARS", base=Currency.XLM, quote=Currency.ARS)
    XLM_BRL = dict(value="XLMBRL", base=Currency.XLM, quote=Currency.BRL)
    XLM_CLP = dict(value="XLMCLP", base=Currency.XLM, quote=Currency.CLP)
    XLM_EUR = dict(value="XLMEUR", base=Currency.XLM, quote=Currency.EUR)
    XLM_MXN = dict(value="XLMMXN", base=Currency.XLM, quote=Currency.MXN)


class TimeFrame(_Enum):
    DAYS_1 = 1440
    HOURS_1 = 60
    MINUTES_15 = 15
    MINUTES_5 = 5
    MINUTES_1 = 1

    @staticmethod
    def _format_value(value):
        values = {
            1440: "DAYS_1",
            60: "HOURS_1",
            15: "MINUTES_15",
            5: "MINUTES_5",
            1: "MINUTES_1",
        }
        return values[value]



import math
from collections import namedtuple
from datetime import datetime


def parse_datetime(datetime_str):
    if datetime_str:
        return datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%fZ")


def int_or_none(value):
    if value:
        return int(value)


def float_or_none(value):
    if value:
        return float(value)


class Amount(
    namedtuple(
        "amount",
        [
            "amount",
            "currency",
        ],
    )
):
    @classmethod
    def create_from_json(cls, amount):
        if amount:
            amount = cls(
                amount=float(amount[0]),
                currency=amount[1],
            )
        return amount


class PagesMeta(
    namedtuple(
        "meta",
        [
            "current_page",
            "total_count",
            "total_pages",
        ],
    )
):
    @classmethod
    def create_from_json(cls, meta):
        if meta:
            return cls(
                current_page=int(meta["current_page"]),
                total_count=int(meta["total_count"]),
                total_pages=int(meta["total_pages"]),
            )
        return meta


class Market(
    namedtuple(
        "market",
        [
            "id",
            "name",
            "base_currency",
            "quote_currency",
            "minimum_order_amount",
            "json",
        ],
    ),
):
    @classmethod
    def create_from_json(cls, market):
        return cls(
            id=market["id"],
            name=market["name"],
            base_currency=market["base_currency"],
            quote_currency=market["quote_currency"],
            minimum_order_amount=Amount.create_from_json(
                market["minimum_order_amount"]
            ),
            json=market,
        )


class Ticker(
    namedtuple(
        "ticker",
        [
            "last_price",
            "min_ask",
            "max_bid",
            "volume",
            "price_variation_24h",
            "price_variation_7d",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, ticker):
        return cls(
            last_price=Amount.create_from_json(ticker["last_price"]),
            min_ask=Amount.create_from_json(ticker["min_ask"]),
            max_bid=Amount.create_from_json(ticker["max_bid"]),
            volume=Amount.create_from_json(ticker["volume"]),
            price_variation_24h=float_or_none(ticker["price_variation_24h"]),
            price_variation_7d=float_or_none(ticker["price_variation_7d"]),
            json=ticker,
        )


class Quotation(
    namedtuple(
        "quotation",
        [
            "amount",
            "base_balance_change",
            "base_exchanged",
            "fee",
            "incomplete",
            "limit",
            "order_amount",
            "quote_balance_change",
            "quote_exchanged",
            "type",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, quotation):
        return cls(
            amount=Amount.create_from_json(quotation["amount"]),
            base_balance_change=Amount.create_from_json(
                quotation["base_balance_change"]
            ),
            base_exchanged=Amount.create_from_json(quotation["base_exchanged"]),
            fee=Amount.create_from_json(quotation["fee"]),
            incomplete=quotation["incomplete"],
            limit=Amount.create_from_json(quotation["limit"]),
            order_amount=Amount.create_from_json(quotation["order_amount"]),
            quote_balance_change=Amount.create_from_json(
                quotation["quote_balance_change"]
            ),
            quote_exchanged=Amount.create_from_json(quotation["quote_exchanged"]),
            type=quotation["type"],
            json=quotation,
        )


class OrderBookEntry(
    namedtuple(
        "book_entry",
        [
            "price",
            "amount",
        ],
    )
):
    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=float(book_entry[0]),
            amount=float(book_entry[1]),
        )


class OrderBook(
    namedtuple(
        "order_book",
        [
            "asks",
            "bids",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, order_book):
        return cls(
            asks=[
                OrderBookEntry.create_from_json(entry) for entry in order_book["asks"]
            ],
            bids=[
                OrderBookEntry.create_from_json(entry) for entry in order_book["bids"]
            ],
            json=order_book,
        )


class FeePercentage(
    namedtuple(
        "fee_percentage",
        [
            "value",
        ],
    )
):
    @classmethod
    def create_from_json(cls, fee_percentage):
        return cls(
            value=float(fee_percentage["value"]),
        )


class Balance(
    namedtuple(
        "balance",
        [
            "id",
            "account_id",
            "amount",
            "available_amount",
            "frozen_amount",
            "pending_withdraw_amount",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, balance):
        return cls(
            id=balance["id"],
            account_id=balance["account_id"],
            amount=Amount.create_from_json(balance["amount"]),
            available_amount=Amount.create_from_json(balance["available_amount"]),
            frozen_amount=Amount.create_from_json(balance["frozen_amount"]),
            pending_withdraw_amount=Amount.create_from_json(
                balance["pending_withdraw_amount"]
            ),
            json=balance,
        )


class Order(
    namedtuple(
        "order",
        [
            "id",
            "account_id",
            "amount",
            "created_at",
            "fee_currency",
            "limit",
            "market_id",
            "original_amount",
            "paid_fee",
            "price_type",
            "state",
            "total_exchanged",
            "traded_amount",
            "type",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, order):
        return cls(
            id=order["id"],
            account_id=order["account_id"],
            amount=Amount.create_from_json(order["amount"]),
            created_at=parse_datetime(order["created_at"]),
            fee_currency=order["fee_currency"],
            limit=Amount.create_from_json(order["limit"]),
            market_id=order["market_id"],
            original_amount=Amount.create_from_json(order["original_amount"]),
            paid_fee=Amount.create_from_json(order["paid_fee"]),
            price_type=order["price_type"],
            state=order["state"],
            total_exchanged=Amount.create_from_json(order["total_exchanged"]),
            traded_amount=Amount.create_from_json(order["traded_amount"]),
            type=order["type"],
            json=order,
        )


class OrderPages(
    namedtuple(
        "order_pages",
        [
            "orders",
            "meta",
        ],
    )
):
    @classmethod
    def create_from_json(cls, orders, pages_meta):
        return cls(
            orders=[Order.create_from_json(order) for order in orders],
            meta=PagesMeta.create_from_json(pages_meta),
        )


class BalanceEvent(
    namedtuple(
        "balance_event",
        [
            "id",
            "account_id",
            "created_at",
            "currency",
            "event",
            "event_ids",
            "new_amount",
            "new_available_amount",
            "new_frozen_amount",
            "new_frozen_for_fee",
            "new_pending_withdraw_amount",
            "old_amount",
            "old_available_amount",
            "old_frozen_amount",
            "old_frozen_for_fee",
            "old_pending_withdraw_amount",
            "transaction_type",
            "transfer_description",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, event):
        return cls(
            id=event["id"],
            account_id=event["account_id"],
            created_at=parse_datetime(event["created_at"]),
            currency=event["currency"],
            event=event["event"],
            event_ids=event["event_ids"],
            new_amount=event["new_amount"],
            new_available_amount=event["new_available_amount"],
            new_frozen_amount=event["new_frozen_amount"],
            new_frozen_for_fee=event["new_frozen_for_fee"],
            new_pending_withdraw_amount=event["new_pending_withdraw_amount"],
            old_amount=event["old_amount"],
            old_available_amount=event["old_available_amount"],
            old_frozen_amount=event["old_frozen_amount"],
            old_frozen_for_fee=event["old_frozen_for_fee"],
            old_pending_withdraw_amount=event["old_pending_withdraw_amount"],
            transaction_type=event["transaction_type"],
            transfer_description=event["transfer_description"],
            json=event,
        )


class BalanceEventPages(
    namedtuple(
        "event_pages",
        [
            "balance_events",
            "meta",
        ],
    )
):
    @classmethod
    def create_from_json(cls, events, total_count, page):
        return cls(
            balance_events=[BalanceEvent.create_from_json(event) for event in events],
            meta=PagesMeta(
                current_page=page or 1,
                total_count=total_count,
                total_pages=math.ceil(total_count / len(events)),
            ),
        )


class TradeTransaction(
    namedtuple(
        "trade_transaction",
        [
            "id",
            "market_id",
            "created_at",
            "updated_at",
            "amount_sold",
            "price_paid",
            "ask_order",
            "bid_order",
            "triggering_order",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, transaction):
        return cls(
            id=transaction["id"],
            market_id=transaction["market_id"],
            created_at=parse_datetime(transaction["created_at"]),
            updated_at=parse_datetime(transaction["updated_at"]),
            amount_sold=Amount.create_from_json(
                [transaction["amount_sold"] / 1e8, transaction["amount_sold_currency"]]
            ),
            price_paid=Amount.create_from_json(
                [transaction["price_paid"] / 1e2, transaction["price_paid_currency"]]
            ),
            ask_order=Order.create_from_json(transaction["ask"]),
            bid_order=Order.create_from_json(transaction["bid"]),
            triggering_order=Order.create_from_json(transaction["triggering_order"]),
            json=transaction,
        )


class TradeTransactionPages(
    namedtuple(
        "trade_transaction_pages",
        [
            "trade_transactions",
            "meta",
        ],
    )
):
    @classmethod
    def create_from_json(cls, transactions, pages_meta):
        return cls(
            trade_transactions=[
                TradeTransaction.create_from_json(transaction)
                for transaction in transactions
            ],
            meta=PagesMeta.create_from_json(pages_meta),
        )


class TradeEntry(
    namedtuple(
        "trade_transaction_pages",
        [
            "timestamp",
            "amount",
            "price",
            "direction",
        ],
    )
):
    @classmethod
    def create_from_json(cls, entry):
        return cls(
            timestamp=int(entry[0]),
            amount=float(entry[1]),
            price=float(entry[2]),
            direction=entry[3],
        )


class Trades(
    namedtuple(
        "trade_transaction_pages",
        [
            "timestamp",
            "last_timestamp",
            "entries",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, trades):
        return cls(
            timestamp=int_or_none(trades["timestamp"]),
            last_timestamp=int_or_none(trades["last_timestamp"]),
            entries=[TradeEntry.create_from_json(entry) for entry in trades["entries"]],
            json=trades,
        )


class TransferData(
    namedtuple(
        "transfer",
        [
            "type",
            "address",
            "tx_hash",
        ],
    )
):
    @classmethod
    def create_from_json(cls, transfer, address_key):
        if transfer:
            return cls(
                type=transfer["type"],
                address=transfer.get(address_key),
                tx_hash=transfer.get("tx_hash"),
            )


class Transfer(
    namedtuple(
        "transfer",
        [
            "id",
            "created_at",
            # 'updated_at', Missing from response?
            "amount",
            "fee",
            "currency",
            "state",
            "data",
            "json",
        ],
    )
):
    address_key = None
    data_key = None

    @classmethod
    def create_from_json(cls, transfer):
        return cls(
            id=transfer["id"],
            created_at=parse_datetime(transfer.get("created_at")),
            amount=Amount.create_from_json(transfer["amount"]),
            # Fee is only returned on withdrawals
            fee=Amount.create_from_json(transfer.get("fee")),
            currency=transfer["currency"],
            state=transfer["state"],
            data=TransferData.create_from_json(transfer[cls.data_key], cls.address_key),
            json=transfer,
        )


class Withdrawal(Transfer):
    address_key = "target_address"
    data_key = "withdrawal_data"


class Deposit(Transfer):
    address_key = "address"
    data_key = "deposit_data"


class WithdrawalPages(
    namedtuple(
        "withdrawal_pages",
        [
            "withdrawals",
            "meta",
        ],
    )
):
    @classmethod
    def create_from_json(cls, withdrawals, pages_meta):
        return cls(
            withdrawals=[
                Withdrawal.create_from_json(withdrawal) for withdrawal in withdrawals
            ],
            meta=PagesMeta.create_from_json(pages_meta),
        )


class DepositPages(
    namedtuple(
        "deposit_pages",
        [
            "deposits",
            "meta",
        ],
    )
):
    @classmethod
    def create_from_json(cls, deposits, pages_meta):
        return cls(
            deposits=[Deposit.create_from_json(deposit) for deposit in deposits],
            meta=PagesMeta.create_from_json(pages_meta),
        )


class AveragePrice(
    namedtuple(
        "reports",
        [
            "datetime",
            "amount",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, report):

        return cls(
            datetime=report[0],
            amount=report[1],
            json=report,
        )


class Candlestick(
    namedtuple(
        "report",
        [
            "datetime",
            "open",
            "high",
            "low",
            "close",
            "volume",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, report):

        return cls(
            datetime=report[0],
            open=report[1],
            high=report[2],
            low=report[3],
            close=report[4],
            volume=report[5],
            json=report,
        )



from collections import namedtuple
from datetime import datetime


def parse_datetime(datetime_str):
    if datetime_str:
        return datetime.strptime(datetime_str, "%Y-%m-%d %H:%M")


def parse_iso_datetime(datetime_str):
    if datetime_str:
        return datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%f")


def check_null(value):
    return value if value != "null" else None


def int_or_null(value):
    value = check_null(value)
    return int(value) if value else None


def float_or_null(value):
    value = check_null(value)
    return float(value) if value else None


class Pagination(namedtuple("pagination", ["previous", "limit", "page", "next"])):
    @classmethod
    def create_from_json(cls, meta):
        if meta:
            return cls(
                previous=int_or_null(meta["previous"]),
                limit=int(meta["limit"]),
                page=int(meta["page"]),
                next=int_or_null(meta["next"]),
            )
        return meta


class Ticker(
    namedtuple(
        "ticker",
        [
            "high",
            "low",
            "ask",
            "bid",
            "last_price",
            "volume",
            "market",
            "timestamp",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, ticker):
        ticker = ticker[0]
        return cls(
            high=float(ticker["high"]),
            low=float(ticker["low"]),
            ask=float(ticker["ask"]),
            bid=float(ticker["bid"]),
            last_price=float(ticker["last_price"]),
            volume=float(ticker["volume"]),
            market=ticker["market"],
            timestamp=parse_iso_datetime(ticker["timestamp"]),
            json=ticker,
        )


class OrderBookEntry(
    namedtuple(
        "book_entry",
        [
            "price",
            "amount",
            "timestamp",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=float(book_entry["price"]),
            amount=float(book_entry["amount"]),
            timestamp=parse_iso_datetime(book_entry["timestamp"]),
            json=book_entry,
        )


class OrderBook(
    namedtuple(
        "order_book",
        [
            "order_book",
            "pagination",
        ],
    )
):
    @classmethod
    def create_from_json(cls, order_book, pagination):
        return cls(
            order_book=[
                OrderBookEntry.create_from_json(book_entry) for book_entry in order_book
            ],
            pagination=Pagination.create_from_json(pagination),
        )


class TradesEntry(
    namedtuple(
        "trades_entry",
        [
            "market_taker",
            "timestamp",
            "price",
            "amount",
            "market",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, trades_entry):
        return cls(
            market_taker=trades_entry["market_taker"],
            timestamp=parse_iso_datetime(trades_entry["timestamp"]),
            price=float(trades_entry["price"]),
            amount=float(trades_entry["amount"]),
            market=trades_entry["market"],
            json=trades_entry,
        )


class Trades(
    namedtuple(
        "trades",
        [
            "trades",
            "pagination",
        ],
    )
):
    @classmethod
    def create_from_json(cls, trades, pagination):
        return cls(
            trades=[
                TradesEntry.create_from_json(trades_entry) for trades_entry in trades
            ],
            pagination=Pagination.create_from_json(pagination),
        )


class WalletBalance(
    namedtuple(
        "wallet_balance",
        [
            "available",
            "balance",
            "wallet",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, balance):
        return cls(
            available=float(balance["available"]),
            balance=float(balance["balance"]),
            wallet=balance["wallet"],
            json=balance,
        )


class Balance(
    namedtuple(
        "balance",
        [
            "ARS",
            "CLP",
            "ETH",
        ],
    )
):
    @classmethod
    def create_from_json(cls, balance):
        return cls(
            ARS=cls.get_wallet_balance(balance, "ARS"),
            CLP=cls.get_wallet_balance(balance, "CLP"),
            ETH=cls.get_wallet_balance(balance, "ETH"),
        )

    @staticmethod
    def get_wallet_balance(balance, wallet: str):
        wallet = [b for b in balance if b["wallet"] == wallet]
        return WalletBalance.create_from_json(wallet[0]) if wallet else None


class OrderAmount(
    namedtuple(
        "wallet_balance",
        [
            "original",
            "remaining",
            "executed",
        ],
    )
):
    @classmethod
    def create_from_json(cls, amount):
        return cls(
            original=float(amount["original"]),
            remaining=float_or_null(amount.get("remaining")),
            executed=float_or_null(amount.get("executed")),
        )


class Order(
    namedtuple(
        "order",
        [
            "id",
            "status",
            "type",
            "price",
            "amount",
            "execution_price",
            "avg_execution_price",
            "market",
            "created_at",
            "updated_at",
            "executed_at",
            "json",
        ],
    )
):
    @classmethod
    def create_from_json(cls, order):
        return cls(
            # Order ID
            id=order["id"],
            # Order status, 'active' or 'executed'
            status=order["status"],
            # Order type, 'buy' or 'sell'
            type=order["type"],
            # Order limit price
            price=float(order["price"]),
            # Order amount
            amount=OrderAmount.create_from_json(order["amount"]),
            # Order execution price
            execution_price=float_or_null(order.get("execution_price")),
            # Order weighted average execution, 0 if not executed
            avg_execution_price=float_or_null(order.get("avg_execution_price")),
            # Market pair
            market=order["market"],
            # Order creation timestamp
            created_at=parse_iso_datetime(order["created_at"]),
            # Order update timestamp. Only on active orders
            updated_at=parse_iso_datetime(order.get("created_at")),
            # Order execution timestamp. Only on executed orders
            executed_at=parse_iso_datetime(order.get("executed_at")),
            # Order JSON data
            json=order,
        )


class Orders(
    namedtuple(
        "orders",
        [
            "orders",
            "pagination",
        ],
    )
):
    @classmethod
    def create_from_json(cls, orders, pagination):
        return cls(
            orders=[Order.create_from_json(order) for order in orders],
            pagination=Pagination.create_from_json(pagination),
        )



from requests import PreparedRequest as P

from .client_public import BitstampPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class BitstampHMACAuth(HMACAuth):

    signature_delimiter = ""

    def __init__(self, api_key: str, secret: str, customer_id: (str, int), **kwargs):
        super().__init__(api_key, secret, **kwargs)
        self.customer_id = str(customer_id)

    def add_api_key(self, r: P):
        r.body["key"] = self.api_key

    def add_nonce(self, r: P, nonce: str):
        r.body["nonce"] = nonce

    def add_signature(self, r: P, nonce: str):
        message = self.build_message(r, nonce)
        signature = self.sign(message)
        r.body["signature"] = signature.upper()

    def auth_request(self, r: P, nonce: str):
        r.body = self.parse_data(r.body)
        super().auth_request(r, nonce)
        r.prepare_body(data=r.body, files=None)

    def build_message(self, r: P, nonce: str):
        components = [nonce, self.customer_id, self.api_key]
        message = self.signature_delimiter.join(components)
        return message


class BitstampAuth(BitstampPublic, AuthMixin):
    auth_cls = BitstampHMACAuth

    def __init__(
        self,
        key: str,
        secret: str,
        customer_id: (str, int),
        timeout: int = None,
        **kwargs,
    ):
        super().__init__(timeout, **kwargs)
        self.add_auth(key, secret, customer_id)

    # Private user data -------------------------------------------------------
    def account_balance(self, currency_pair: str = None):
        """
        Returns amounts available and reserved per currency,
        also fee percentage for the currency pair, or for all currency pairs
        if no pair was specified:

            {'btc_reserved': '0',
             'fee': '0.5000',
             'btc_available': '2.30856098',
             'usd_reserved': '0',
             'btc_balance': '2.30856098',
             'usd_balance': '114.64',
             'usd_available': '114.64',
             ...
             ---If currency pair was specified:
             'fee': '',
             ---If currency pair was not specified:
             'btcusd_fee': '0.25',
             'btceur_fee': '0.25',
             'eurusd_fee': '0.20',
             ...
             }
        """
        endpoint = self._endpoint_for(
            f"balance/{currency_pair}" if currency_pair else "balance"
        )
        return self.post(endpoint)

    def user_transactions(
        self,
        currency_pair: str = None,
        offset: int = None,
        limit: int = None,
        sort_desc: bool = None,
    ):
        """
        Returns descending list of transactions. Every transaction (dictionary)
        contains:

            {'usd': '-39.25',
             'datetime': '2013-03-26 18:49:13',
             'fee': '0.20',
             'btc': '0.50000000',
             'type': 2,
             'id': 213642}

        Instead of the keys btc and usd, it can contain other currency codes
        """
        payload = {
            "offset": offset,
            "limit": limit,
        }
        if sort_desc is not None:
            payload["sort"] = "desc" if sort_desc else "asc"
        endpoint = self._endpoint_for("user_transactions")
        if currency_pair:
            endpoint = f"{endpoint}/{currency_pair}/"
        return self.post(endpoint, data=payload)

    # Orders ------------------------------------------------------------------
    def open_orders(self, currency_pair: str = None):
        """
        Returns JSON list of open orders. Each order is represented as a
        dictionary.
        """
        endpoint = self._endpoint_for(f"open_orders/{currency_pair or 'all'}")
        return self.post(endpoint)

    def orders_status(self, order_id: int):
        """
        Returns dictionary.
        - status: 'Finished'
          or      'In Queue'
          or      'Open'
        - transactions: list of transactions
          Each transaction is a dictionary with the following keys:
              btc, usd, price, type, fee, datetime, tid
          or  btc, eur, ....
          or  eur, usd, ....
        """
        endpoint = self._endpoint_for("order_status", version=1)
        return self.post(endpoint, data={"id": order_id})

    def cancel_order(self, order_id: int):
        """
        Cancel the order specified by order_id.

        Returns dictionary of the canceled order, containing the keys:
        id, type, price, amount
        """
        endpoint = self._endpoint_for("cancel_order")
        return self.post(endpoint, data={"id": order_id})

    def cancel_all_orders(self):
        """
        Cancel all open orders.

        Returns True if it was successful
        """
        endpoint = self._endpoint_for("cancel_all_orders", version=1)
        return self.post(endpoint)

    def _limit_order(
        self,
        endpoint,
        currency_pair: str,
        amount: float,
        price: float,
        limit_price: float,
        daily_order: bool = None,
    ):
        endpoint = self._endpoint_for(endpoint % currency_pair)
        return self.post(
            endpoint,
            data={
                "amount": amount,
                "price": price,
                "limit_price": limit_price,
                "daily_order": daily_order,
            },
        )

    def _market_order(self, endpoint: str, currency_pair: str, amount: float):
        endpoint = self._endpoint_for(endpoint % currency_pair)
        return self.post(endpoint, data={"amount": amount})

    def buy_limit_order(
        self,
        currency_pair: str,
        amount: float,
        price: float,
        limit_price: float,
        daily_order: bool = None,
    ):
        """
        Order to buy amount for specified price.
        """
        return self._limit_order(
            "buy/%s", currency_pair, amount, price, limit_price, daily_order
        )

    def sell_limit_order(
        self,
        currency_pair: str,
        amount: float,
        price: float,
        limit_price: float,
        daily_order: bool = None,
    ):
        """
        Order to sell amount for specified price.
        """
        return self._limit_order(
            "sell/%s", currency_pair, amount, price, limit_price, daily_order
        )

    def buy_market_order(self, currency_pair: str, amount: float):
        """
        Order to buy amount for market price.
        """
        return self._market_order("buy/market/%s", currency_pair, amount)

    def sell_market_order(self, currency_pair: str, amount: float):
        """
        Order to sell amount for market price.
        """
        return self._market_order("sell/market/%s", currency_pair, amount)

    # Withdrawals -------------------------------------------------------------
    def withdrawal_requests(self, time_delta: int = None):
        """
        Returns list of withdrawal requests.

        Each request is represented as a dictionary.

        By default, the last 24 hours (86400 seconds) are returned.
        """
        endpoint = self._endpoint_for("withdrawal-requests")
        return self.post(endpoint, data={"timedelta": time_delta})

    def withdrawal(
        self, method: str, address: str, amount: float, version: int = 2, **kwargs
    ):
        endpoint = self._endpoint_for(f"{method}_withdrawal", version)
        return self.post(
            endpoint, data={"amount": amount, "address": address, **kwargs}
        )

    def bitcoin_withdrawal(self, address: str, amount: float, instant: bool = None):
        """
        Send bitcoin to another bitcoin wallet specified by address.
        """
        return self.withdrawal(
            "bitcoin",
            address,
            amount,
            **{
                "version": 1,
                "instant": (1 if instant else 0) if instant else None,
            },
        )

    def ripple_withdrawal(self, address: str, amount: float, currency: str):
        """
        Returns true if successful.
        """
        return self.withdrawal("ripple", address, amount, version=1, currency=currency)

    def litecoin_withdrawal(self, address: str, amount: float):
        """
        Send litecoin to another litecoin wallet specified by address.
        """
        return self.withdrawal("ltc", address, amount)

    def eth_withdrawal(self, address: str, amount: float):
        """
        Send ether to another ether wallet specified by address.
        """
        return self.withdrawal("eth", address, amount)

    def bch_withdrawal(self, address: str, amount: float):
        """
        Send bitcoin cash to another bitcoin cash wallet specified by address.
        """
        return self.withdrawal("bch", address, amount)

    def xrp_withdrawal(self, address: str, amount: float, destination_tag: str = None):
        """
        Sends xrp to another xrp wallet specified by address.
        Returns withdrawal id.
        """
        return self.withdrawal("xrp", address, amount, destination_tag=destination_tag)

    # Deposits ----------------------------------------------------------------
    def deposit_address(self, method: str, version: int = 2):
        endpoint = self._endpoint_for(f"{method}_address", version)
        return self.post(endpoint)

    def bitcoin_deposit_address(self):
        """
        Returns bitcoin deposit address as unicode string
        """
        return self.deposit_address("bitcoin_deposit", version=1)

    def ripple_deposit_address(self):
        """
        Returns ripple deposit address as unicode string.
        """
        return self.deposit_address("ripple", version=1)

    def litecoin_deposit_address(self):
        """
        Returns litecoin deposit address as unicode string
        """
        return self.deposit_address("ltc")

    def eth_deposit_address(self):
        """
        Returns ethereum deposit address as unicode string
        """
        return self.deposit_address("eth")

    def bch_deposit_address(self):
        """
        Returns bitcoin cash deposit address as unicode string
        """
        return self.deposit_address("bch")

    def xrp_deposit_address(self):
        """
        Returns ripple deposit address and destination tag as dictionary.
        Example:
            {'destination_tag': 53965834,
             'address': 'rDsBeAmAa4FFwbQTJp9Rs84Q56vCiWCaBx'}
        """
        return self.deposit_address("xrp")

    def unconfirmed_bitcoin_deposits(self):
        """
        Returns JSON list of unconfirmed bitcoin transactions.
        Each transaction is represented as dictionary:

        amount
          bitcoin amount
        address
          deposit address used
        confirmations
          number of confirmations
        """
        endpoint = self._endpoint_for("unconfirmed_btc", version=1)
        return self.post(endpoint)

    # Transfers ---------------------------------------------------------------
    def _transfer(
        self, endpoint: str, amount: float, currency: str, sub_account: int = None
    ):
        endpoint = self._endpoint_for(endpoint)
        return self.post(
            endpoint,
            data={
                "amount": amount,
                "currency": currency,
                "subAccount": sub_account,
            },
        )

    def transfer_to_main(self, amount: float, currency: str, sub_account: int = None):
        """
        Returns dictionary with status.
        sub account has to be the numerical id of the sub account, not the name
        """
        return self._transfer("transfer-to-main", amount, currency, sub_account)

    def transfer_from_main(self, amount: float, currency: str, sub_account: int):
        """
        Returns dictionary with status.
        sub account has to be the numerical id of the sub account, not the name
        """
        return self._transfer("transfer-from-main", amount, currency, sub_account)

    # TODO: Bank methods



import base64
import hashlib
import hmac

from requests import PreparedRequest as P

from .client_public import KrakenPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class KrakenHMACAuth(HMACAuth):

    api_key_header = "API-Key"
    signature_header = "API-Sign"
    algorithm = "sha512"

    def add_nonce(self, r: P, nonce: str):
        body = self.parse_data(r.body)
        body["nonce"] = nonce
        r.prepare_body(data=body, files=None)

    def build_message(self, r: P, nonce: str):
        auth = (nonce + r.body).encode()
        digest = hashlib.sha256(auth).digest()
        message = r.path_url.encode() + digest
        return message

    def sign(self, msg: str):

        h = hmac.new(
            key=base64.b64decode(self.secret), msg=msg, digestmod=self.algorithm
        )

        signature = base64.b64encode(h.digest()).decode()

        return signature


class KrakenAuth(KrakenPublic, AuthMixin):
    auth_cls = KrakenHMACAuth

    def __init__(self, key: str, secret: str, timeout: int = None, **kwargs):
        super().__init__(timeout, **kwargs)
        self.add_auth(key, secret)

    # Private user data -------------------------------------------------------
    # Get account balance.
    def balance(self):
        return self.post("private/Balance")

    # Get trade balance.
    def trade_balance(self, asset: str = None, asset_class: str = None):
        return self.post(
            "private/TradeBalance",
            data={
                "asset": str(asset) if asset else None,
                "aclass": str(asset_class) if asset_class else None,
            },
        )

    # Get open orders.
    def open_orders(self, include_trades: bool = None, userref: str = None):
        return self.post(
            "private/OpenOrders",
            data={
                "trades": include_trades,
                "userref": userref,
            },
        )

    # Get closed orders.
    def closed_orders(
        self,
        include_trades: bool = None,
        userref: str = None,
        start: int = None,
        end: int = None,
        ofs: int = None,
        closetime: int = None,
    ):
        return self.post(
            "private/ClosedOrders",
            data={
                "trades": include_trades,
                "userref": userref,
                "start": start,
                "end": end,
                "ofs": ofs,
                "closetime": closetime,
            },
        )

    # Query orders info.
    def query_orders(
        self, txid: list, include_trades: bool = None, userref: str = None
    ):
        return self.post(
            "private/QueryOrders",
            data={
                "txid": txid,
                "trades": include_trades,
                "userref": userref,
            },
        )

    # Get trades history.
    def trades_history(
        self,
        trade_type: str = None,
        include_trades: bool = None,
        start: int = None,
        end: int = None,
        ofs: int = None,
    ):
        return self.post(
            "private/TradesHistory",
            data={
                "type": str(trade_type) if trade_type else None,
                "trades": include_trades,
                "start": start,
                "end": end,
                "ofs": ofs,
            },
        )

    # Query trades info.
    def query_trades(self, txid: list, include_trades: bool = None):
        return self.post(
            "private/QueryTrades",
            data={
                "txid": txid,
                "trades": include_trades,
            },
        )

    # Query trades info.
    def open_positions(self, txid: list = None, include_pl: bool = None):
        return self.post(
            "private/OpenPositions",
            data={
                "txid": txid,
                "docalcs": include_pl,
            },
        )

    # Get ledgers info.
    def ledgers(
        self,
        asset_class: str = None,
        asset: str = None,
        ledger_type: str = None,
        start: int = None,
        end: int = None,
        ofs: int = None,
    ):
        return self.post(
            "private/Ledgers",
            data={
                "aclass": str(asset_class) if asset_class else None,
                "asset": str(asset) if asset else None,
                "type": str(ledger_type) if ledger_type else None,
                "start": start,
                "end": end,
                "ofs": ofs,
            },
        )

    # Query ledgers.
    def query_ledgers(self, ledger_id: str):
        return self.post(
            "private/QueryLedgers",
            data={
                "id": ledger_id,
            },
        )

    # Get trade volume.
    def trade_volume(self, pair: str = None, fee_info: bool = None):
        return self.post(
            "private/TradeVolume",
            data={
                "pair": str(pair) if pair else None,
                "fee-info": fee_info,
            },
        )

    # Private user trading  ---------------------------------------------------
    # Add standard order
    def add_order(
        self,
        pair: str,
        direction: str,
        order_type: str,
        volume: float,
        price: float = None,
        price2: float = None,
        leverage: float = None,
        oflags: list = None,
        starttm: int = None,
        expiretm: int = None,
        userref: str = None,
        validate: bool = None,
        c_ordertype: str = None,
        c_price: float = None,
        c_price2: float = None,
    ):
        return self.post(
            "private/AddOrder",
            data={
                "pair": str(pair),
                "type": str(direction),
                "ordertype": str(order_type),
                "price": price,
                "price2": price2,
                "volume": volume,
                "leverage": leverage,
                "oflags": oflags,
                "starttm": starttm,
                "expiretm": expiretm,
                "userref": userref,
                "validate": validate,
                "close[ordertype]": str(c_ordertype) if c_ordertype else None,
                "close[price]": c_price,
                "close[price2]": c_price2,
            },
        )

    # Cancel open order
    def cancel_order(self, txid: str):
        return self.post(
            "private/CancelOrder",
            data={
                "txid": txid,
            },
        )

    # Private user funding  ---------------------------------------------------
    # Get deposit methods.
    def deposit_methods(self, asset: str, asset_class: str = None):
        return self.post(
            "private/DepositMethods",
            data={
                "asset": str(asset),
                "aclass": str(asset_class) if asset_class else None,
            },
        )

    # Get deposit addresses
    def deposit_addresses(
        self, asset: str, method: str, asset_class: str = None, new: bool = None
    ):
        return self.post(
            "private/DepositAddresses",
            data={
                "asset": str(asset),
                "method": str(method),
                "aclass": str(asset_class) if asset_class else None,
                "new": new,
            },
        )

    # Get status of recent deposits
    def deposit_status(self, asset: str, method: str, asset_class: str = None):
        return self.post(
            "private/DepositStatus",
            data={
                "asset": str(asset),
                "method": str(method),
                "aclass": str(asset_class) if asset_class else None,
            },
        )

    # Get withdrawal information
    def withdraw_info(
        self, asset: str, amount: float, key: str, asset_class: str = None
    ):
        return self.post(
            "private/WithdrawInfo",
            data={
                "asset": str(asset),
                "aclass": str(asset_class) if asset_class else None,
                "amount": amount,
                "key": key,
            },
        )

    # Withdraw funds
    def withdraw(self, asset: str, amount: float, key: str, asset_class: str = None):
        return self.post(
            "private/Withdraw",
            data={
                "asset": str(asset),
                "aclass": str(asset_class) if asset_class else None,
                "amount": amount,
                "key": key,
            },
        )

    # Get status of recent withdrawals
    def withdraw_status(self, asset: str, method: str, asset_class: str = None):
        return self.post(
            "private/WithdrawStatus",
            data={
                "asset": str(asset),
                "method": str(method),
                "aclass": str(asset_class) if asset_class else None,
            },
        )

    # Request withdrawal cancellation
    def withdraw_cancel(self, asset: str, refid: str, asset_class: str = None):
        return self.post(
            "private/WithdrawCancel",
            data={
                "asset": str(asset),
                "aclass": str(asset_class) if asset_class else None,
                "refid": refid,
            },
        )



from . import models as _m
from ..base import Client, ModelMixin


class RipioExchangePublic(Client, ModelMixin):
    """API Doc: https://www.ripio.com/docs/#ripio-exchange"""

    base_url = "https://exchange.ripio.com/api/v1/"
    error_keys = ["detail"]

    def order_books(self):
        """Fetch order books for all markets"""
        data = self.get("book/")
        if self.return_json:
            return data
        return {
            market: _m.OrderBook.create_from_json(book) for market, book in data.items()
        }

    def order_book(self, market: str):
        """Fetch order book for the provided market"""
        data = self.order_books()
        return data[market]


# TODO: Ripio Auth not implemented
class RipioExchangeAuth(RipioExchangePublic):
    def trades(self, page: int = None):
        """Fetch last trades"""
        data = self.get(
            "trades/",
            params={
                "page": page,
            },
        )
        if self.return_json:
            return data
        return _m.Trades.create_from_json(data)


class RipioPublic(Client, ModelMixin):
    """API Doc: https://www.ripio.com/docs/#transactions"""

    base_url = "https://ripio.com/api/v1/"
    error_keys = ["detail"]

    def __init__(self, timeout: int = None, **kwargs):
        super().__init__(timeout, **kwargs)
        self.exchange = RipioExchangePublic(timeout, **kwargs)

    def rates_raw(self):
        return self.get("rates/")

    def rates(self):
        """Fetch rates"""
        data = self.rates_raw()
        data = {"base": data["base"], "rates": data["rates"]}
        if self.return_json:
            return data
        return _m.Rates.create_from_json(data)

    def variation(self):
        """Fetch rates variation"""
        data = self.rates_raw()
        return data["variation"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['trading_api_wrappers',
 'trading_api_wrappers.bitcoinity',
 'trading_api_wrappers.bitex',
 'trading_api_wrappers.bitfinex',
 'trading_api_wrappers.bitstamp',
 'trading_api_wrappers.buda',
 'trading_api_wrappers.coindesk',
 'trading_api_wrappers.coinmarketcap',
 'trading_api_wrappers.cryptomkt',
 'trading_api_wrappers.currencylayer',
 'trading_api_wrappers.kraken',
 'trading_api_wrappers.oxr',
 'trading_api_wrappers.ripio',
 'trading_api_wrappers.sfox',
 'trading_api_wrappers.surbtc']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=1.11.1,<2.0.0',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'requests>=2.26.0,<3.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4.6.4,<5.0.0']}

setup_kwargs = {
    'name': 'trading-api-wrappers',
    'version': '0.18.1',
    'description': 'Clients for popular Crypto Exchanges and other useful services',
    'long_description': '# Trading API Wrappers\n\n> Python 3.6+ clients for popular **Crypto Exchanges** and other useful services.\n\n> **Disclaimer:** Still at an early stage of development. Rapidly evolving APIs.\n\n[![PyPI - License](https://img.shields.io/pypi/l/trading-api-wrappers.svg)](https://opensource.org/licenses/MIT)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/trading-api-wrappers.svg)\n[![PyPI](https://img.shields.io/pypi/v/trading-api-wrappers.svg)](https://pypi.org/project/trading-api-wrappers/)\n![PyPI - Status](https://img.shields.io/pypi/status/trading-api-wrappers.svg)\n[![Updates](https://pyup.io/repos/github/delta575/trading-api-wrappers/shield.svg)](https://pyup.io/repos/github/delta575/trading-api-wrappers/)\n\nSupported APIs:\n\n- [Buda](https://www.buda.com)\n- [Bitfinex](https://www.bitfinex.com)\n- [Bitstamp](https://www.bitstamp.net)\n- [CoinDesk](https://www.coindesk.com)\n- [CoinMarketCap](https://coinmarketcap.com)\n- [CryptoMKT](https://www.cryptomkt.com)\n- [Kraken](https://www.kraken.com)\n- [OpenExchangeRates](https://openexchangerates.org)\n\n## Installation\n\n### Requirements\n\n- Python 3.6+\n\nTo install, simply use `poetry` (or `pip`, of course):\n\n```bash\n$ poetry add trading-api-wrappers\n```\n\n### Dev setup\n\n```bash\n$ poetry install\n```\n\nRename `.env.example` to `.env` and configure your credentials (for tests)\n\n## Usage\n\n### Buda\n\nPublic API:\n\n```python\nfrom trading_api_wrappers import Buda\nclient = Buda.Public()\n```\n\nAuthenticated API:\n\n```python\nfrom trading_api_wrappers import Buda\nclient = Buda.Auth(API_KEY, API_SECRET)\n```\n\nBuda API Doc:\nhttps://api.buda.com\n\n### Bitfinex\n\nPublic API:\n\n```python\nfrom trading_api_wrappers import Bitfinex\nclient = Bitfinex.Public()\n```\n\nAuthenticated API:\n\n```python\nfrom trading_api_wrappers import Bitfinex\nclient = Bitfinex.Auth(API_KEY, API_SECRET)\n```\n\nBitfinex API Doc:\nhttps://bitfinex.readme.io/v1/docs\n\n### Bitstamp\n\nPublic API:\n\n```python\nfrom trading_api_wrappers import Bitstamp\nclient = Bitstamp.Public()\n```\n\nAuthenticated API:\n\n```python\nfrom trading_api_wrappers import Bitstamp\nclient = Bitstamp.Auth(API_KEY, API_SECRET, CUSTOMER_ID)\n```\n\nBitstamp API Doc:\nhttps://www.bitstamp.net/api\n\n### Kraken\n\nPublic API:\n\n```python\nfrom trading_api_wrappers import Kraken\nclient = Kraken.Public()\n```\n\nAuthenticated API:\n\n```python\nfrom trading_api_wrappers import Kraken\nclient = Kraken.Auth(API_KEY, API_SECRET)\n```\n\nKraken API Doc:\nhttps://www.kraken.com/help/api\n\n### CoinDesk\n\n```python\nfrom trading_api_wrappers import CoinDesk\nclient = CoinDesk()\n```\n\nCoinDesk API Doc:\nhttps://www.coindesk.com/api\n\n### CoinMarketCap\n\n```python\nfrom trading_api_wrappers import CoinMarketCap\nclient = CoinMarketCap()\n```\n\nCoinMarketCap API Doc:\nhttps://coinmarketcap.com/api\n\n### CryptoMKT\n\nPublic API:\n\n```python\nfrom trading_api_wrappers import CryptoMKT\nclient = CryptoMKT.Public()\n```\n\nAuthenticated API:\n\n```python\nfrom trading_api_wrappers import CryptoMKT\nclient = CryptoMKT.Auth(API_KEY, API_SECRET)\n```\n\nCryptoMKT API Doc:\nhttps://developers.cryptomkt.com\n\n### OpenExchangeRates\n\n```python\nfrom trading_api_wrappers import OXR\nclient = OXR(APP_ID)\n```\n\nOpenExchangeRates API Doc:\nhttps://docs.openexchangerates.org\n\n### CurrencyLayer\n\n```python\nfrom trading_api_wrappers import CurrencyLayer\nclient = CurrencyLayer(ACCESS_KEY)\n```\n\nCurrencyLayer API Doc:\nhttps://currencylayer.com/documentation\n\n## Licence\n\n[![PyPI - License](https://img.shields.io/pypi/l/trading-api-wrappers.svg)](https://opensource.org/licenses/MIT)\n[![FOSSA Status](https://app.fossa.io/api/projects/git%2Bgithub.com%2Fdelta575%2Ftrading-api-wrappers.svg?type=shield)](https://app.fossa.io/projects/git%2Bgithub.com%2Fdelta575%2Ftrading-api-wrappers?ref=badge_shield)\n\nThe MIT License\n\nCopyright © 2017\n[Felipe Aránguiz](mailto://faranguiz575@gmail.com) | [Sebastián Aránguiz](mailto://sarang575@gmail.com)\n\nSee [LICENSE](LICENSE)\n\n[![FOSSA Status](https://app.fossa.io/api/projects/git%2Bgithub.com%2Fdelta575%2Ftrading-api-wrappers.svg?type=large)](https://app.fossa.io/projects/git%2Bgithub.com%2Fdelta575%2Ftrading-api-wrappers?ref=badge_large)\n\n## Donations\n\nBitcoin:\n\n    186kDw9LFcPvup17YSrWZbFqdZzELUFad3\n\nEther:\n\n    0xeF38fA6c0a37A1BdB60CADd7f6e71F351F6d2583\n',
    'author': 'Felipe Aranguiz',
    'author_email': 'faranguiz575@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/snake575/trading-bots',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



from ..base import Client


class Bitcoinity(Client):
    base_url = "https://bitcoinity.org/"
    timeout = 15

    def ticker(self, currency: str, exchange: str, span: str):
        return self.get(
            "markets/get_ticker",
            params={
                "currency": currency,
                "exchange": exchange,
                "span": span,
            },
        )



from datetime import datetime

from . import constants as _c
from . import models as _m
from ..base import Client, ModelMixin


class CryptoMKTPublic(Client, ModelMixin):
    base_url = "https://api.cryptomkt.com/v1/"
    error_keys = ["message"]

    def markets(self):
        data = self.get("market")
        return data["data"]

    def ticker(self, market_id: str):
        data = self.get("ticker", params={"market": str(market_id)})
        if self.return_json:
            return data
        return _m.Ticker.create_from_json(data["data"])

    def order_book(
        self,
        market_id: str,
        order_type: str,
        page: int = None,
        limit: int = _c.ORDERS_LIMIT,
    ):
        data = self.get(
            "book",
            params={
                "market": str(market_id),
                "type": str(order_type),
                "page": page,
                "limit": limit,
            },
        )
        if self.return_json:
            return data
        return _m.OrderBook.create_from_json(data["data"], data["pagination"])

    def trades(
        self,
        market_id: str,
        start: datetime = None,
        end: datetime = None,
        page: int = None,
        limit: int = _c.ORDERS_LIMIT,
    ):
        if isinstance(start, datetime):
            start = start.strftime("%Y-%m-%d")
        if isinstance(end, datetime):
            end = end.strftime("%Y-%m-%d")
        data = self.get(
            "trades",
            params={
                "market": str(market_id),
                "start": start,
                "end": end,
                "page": page,
                "limit": limit,
            },
        )
        if self.return_json:
            return data
        return _m.Trades.create_from_json(data["data"], data.get("pagination"))



import base64

from requests import PreparedRequest as P

from .client_public_v1 import BitfinexPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class BitfinexHMACAuth(HMACAuth):

    api_key_header = "x-bfx-apikey"
    signature_header = "x-bfx-signature"
    payload_header = "x-bfx-payload"
    algorithm = "sha384"

    def add_nonce(self, r: P, nonce: str):
        pass

    def add_signature(self, r: P, nonce: str):
        message = self.build_message(r, nonce)
        r.headers[self.payload_header] = message
        signature = self.sign(message)
        r.headers[self.signature_header] = signature

    def build_message(self, r: P, nonce: str):
        body = self.load_json(r.body)
        body["request"] = r.path_url
        body["nonce"] = nonce
        encoded_msg = base64.b64encode(self.encode_json(body))
        return encoded_msg


class BitfinexAuth(BitfinexPublic, AuthMixin):
    auth_cls = BitfinexHMACAuth

    def __init__(self, key: str, secret: str, timeout: int = None, **kwargs):
        super().__init__(timeout, **kwargs)
        self.add_auth(key, secret)

    # INFO --------------------------------------------------------------------
    # Return information about your account (trading fees).
    def account_info(self):
        return self.get("account_infos")

    # Return information about your account (trading fees).
    def summary(self):
        return self.get("summary")

    # Check the permissions of the key being used to generate this request.
    def key_info(self):
        return self.get("key_info")

    # See your trading wallet information for margin trading.
    def margin_info(self):
        return self.get("margin_infos")

    # See your balances.
    def balances(self):
        return self.get("balances")

    # MOVEMENTS ---------------------------------------------------------------
    # Return your deposit address to make a new deposit.
    def new_deposit(self, method: str, wallet_name: str, renew: bool = None):
        return self.post(
            "deposit/new",
            json={
                "method": method,
                "wallet_name": wallet_name,
                "renew": renew if renew is not None else None,
            },
        )

    # Allow you to move available balances between your wallets.
    def transfer(self, amount: float, currency: str, wallet_from: str, wallet_to: str):
        return self.post(
            "transfer",
            json={
                "amount": str(amount),
                "currency": str(currency),
                "walletfrom": wallet_from,
                "walletto": wallet_to,
            },
        )

    # Allow you to request a withdrawal from one of your wallet.
    def withdraw(self, w_type: str, wallet: str, amount: float, address: str):
        return self.post(
            "withdraw",
            json={
                "withdraw_type": str(w_type),
                "walletselected": wallet,
                "amount": str(amount),
                "address": address,
            },
        )

    # ORDERS ------------------------------------------------------------------
    # Submit a new order.
    def place_order(
        self,
        amount: float,
        price: float,
        side: str,
        ord_type: str,
        symbol: str,
        params: dict = None,
    ):
        payload = {
            "symbol": str(symbol),
            "amount": str(amount),
            "price": str(price),
            "side": str(side),
            "type": str(ord_type),
            "ocoorder": False,
        }
        payload.update(params or {})
        return self.post("order/new", json=payload)

    # Submit a new order.
    def place_oco_order(
        self,
        amount: float,
        price: float,
        side: str,
        ord_type: str,
        symbol: str,
        buy_price_oco: float,
        sell_price_oco: float,
    ):
        return self.place_order(
            amount,
            price,
            side,
            ord_type,
            symbol,
            params={
                "ocoorder": True,
                "buy_price_oco": str(buy_price_oco),
                "sell_price_oco": str(sell_price_oco),
            },
        )

    # Cancel an order.
    def delete_order(self, order_id: int):
        return self.post("order/cancel", json={"order_id": order_id})

    # Cancel all orders.
    def delete_all_order(self):
        return self.post("order/cancel/all")

    # Get the status of an order. Is it active? Was it cancelled?
    # To what extent has it been executed? etc.
    def status_order(self, order_id: int):
        return self.post("order/status", json={"order_id": order_id})

    # View your active orders.
    def active_orders(self):
        return self.post("orders")

    # View your latest inactive orders.
    # Limited to last 3 days and 1 request per minute.
    def orders_history(self, limit: int):
        return self.post("orders/hist", json={"limit": limit})

    # POSITIONS ---------------------------------------------------------------
    # View your active positions.
    def active_positions(self):
        return self.post("positions")

    # Claim a position.
    def claim_position(self, position_id: int):
        return self.post("position/claim", json={"position_id": position_id})

    # HISTORICAL DATA ---------------------------------------------------------
    # View all of your balance ledger entries.
    def balance_history(
        self,
        currency: str,
        since: float = None,
        until: float = None,
        limit: int = None,
        wallet: str = None,
    ):
        return self.post(
            "history",
            json={
                "currency": str(currency),
                "since": since,
                "until": until,
                "limit": limit,
                "wallet": wallet,
            },
        )

    # View your past deposits/withdrawals.
    def movements(
        self,
        currency: str,
        method: str = None,
        since: float = None,
        until: float = None,
        limit: int = None,
    ):
        return self.post(
            "history/movements",
            json={
                "currency": str(currency),
                "method": method,
                "since": since,
                "until": until,
                "limit": limit,
            },
        )

    # View your past trades.
    def past_trades(
        self,
        symbol: str,
        timestamp: float = None,
        until: float = None,
        limit_trades: int = None,
        reverse: bool = None,
    ):
        return self.post(
            "mytrades",
            json={
                "symbol": str(symbol),
                "timestamp": timestamp,
                "until": until,
                "limit_trades": limit_trades,
                "reverse": reverse if reverse is not None else None,
            },
        )

    # MARGIN FUNDING ----------------------------------------------------------
    # Submit a new Offer.
    def place_offer(
        self, currency: str, amount: float, rate: float, period: int, direction: str
    ):
        return self.post(
            "offer/new",
            json={
                "currency": str(currency),
                "amount": str(amount),
                "rate": str(rate),
                "period": period,
                "direction": str(direction),
            },
        )

    # Cancel an offer.
    def cancel_offer(self, offer_id: int):
        return self.post("offer/cancel", json={"offer_id": offer_id})

    # Get the status of an offer. Is it active? Was it cancelled?
    # To what extent has it been executed? etc.
    def status_offer(self, offer_id: int):
        return self.post("offer/status", json={"offer_id": offer_id})

    # View your active offers.
    def active_offers(self):
        return self.post("offers")
